/*  XmmsTcl - an xmms remote control library for tcl.
 *
 *  Copyright (C) 2002  Pascal Scheffers <pascal@scheffers.net>
 *
 *    $Id: xmmstcl-int.h,v 1.2 2002/03/16 23:27:52 pascal Exp $
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Internal defines, not part of public interface.
 */

#ifndef XMMSTCL_INT_H
#define XMMSTCL_INT_H

#define debug_print(msg) fprintf(stderr, msg)

#ifndef HAVE_GLIB
// This seems prudent, but I have no idea if you can actually compile
// xmmsctrl without GLib installed. 
#define gchar char
#define g_free free
#endif

//messages
#define  E_NO_SESSION_MSG "xmms session is not running"
#define  E_BAD_CMD_MSG "Not a valid xmms command. Usage xmms command ?arg arg ..
. arg?. Valid commands are: isrunning, playlist, version, add, delete, play, pause, stop, isplaying, ispaused, position, length, clear, time, volume, mainvolume, balance, skin, file, title, info, mainwindow, plwindow, eqwindow, showpreferences, alwaysontop, eject, previous, next, addurl, repeat, shuffle, eq, eqpreamp, eqband, quit."
#define  E_NO_IMPL "xmms command not (yet) implemented"
#define  E_BAD_SESSION_ARG_MSG "xmms '-session ?session?' argument not specified"
#define  E_BAD_POS_ARG_MSG "xmms file ?-pos n? argument not specified"
#define  E_BAD_EQ_LIST_MSG "Equalizer list must have 11 float elements"
#define  E_BAD_EQ_BAND_MSG "Equalizer band must be from 0 to 9"
#define  E_NO_MORE_SESSION_MSG "Tried all sessions, but all of them were running"
#define  E_FORK_FAILED_MSG "Could not fork() the new process"
#define  E_NO_MEM_MSG "Not enough memory/Tcl_Alloc failed."
#define  MAX_EXPECTED_SESSIONS 1024
#define  E_EXEC_TIMEOUT_MSG "Timout occured "
#define  E_BAD_TIMEOUT_ARG_MSG "-timeout requires an integer argument"
#define  E_ONCHANGE_USAGE_MSG "Usage: xmms onchange callback_proc ?-cancel -session n?"

#define  MAX_EXEC_WAIT_TIME 5000 //5 seconds in milliseconds 

#define  MIN_TIMEOUT   100
#define  MAX_TIMEOUT   750
#define  TIMEOUT_INCR  10

//xmms ?command? values
#define  CMD_BADCOMMAND        -1
#define  CMD_ISRUNNING         0
#define  CMD_PLAYLIST          1
#define  CMD_GETVERSION        2
#define  CMD_PLAYLISTADD       3
#define  CMD_PLAYLISTDELETE    4
#define  CMD_PLAY              5
#define  CMD_PAUSE             6
#define  CMD_STOP              7
#define  CMD_ISPLAYING         8
#define  CMD_ISPAUSED          9
#define  CMD_PLAYLISTPOS       10 /*set and get */
#define  CMD_PLAYLISTLENGTH    11
#define  CMD_PLAYLISTCLEAR     12
#define  CMD_TIME              13 /* set and get */
#define  CMD_VOLUME            15 /* set and get */
#define  CMD_MAINVOLUME        16 /* set and get */
#define  CMD_BALANCE           17 /* set and get */
#define  CMD_SKIN              18 /* set and get */
#define  CMD_PLAYLISTFILE      19 /* (set and) get */
#define  CMD_PLAYLISTTITLE     20 /* (set and) get */
#define  CMD_INFO              21 /* rate, freq, nchannels */
#define  CMD_MAINWIN           22 /* toggle, (set,) get */
#define  CMD_PLWIN             23 /* toggle, (set,) get */
#define  CMD_EQWIN             24 /* toggle, (set,) get */
#define  CMD_SHOWPREFS         25 
#define  CMD_AOT               26 /* toggle, set, (get) */
#define  CMD_EJECT             27
#define  CMD_PREV              28
#define  CMD_NEXT              29
#define  CMD_ADDURL            30
#define  CMD_REPEAT            31 /* set, (get) */
#define  CMD_SHUFFLE           32 /* set, (get) */
#define  CMD_EQ                33 /* (set, get) */
#define  CMD_EQBAND            34 /* (set, get) */
#define  CMD_EQPREAMP          35 /* (set, get) */
#define  CMD_QUIT              36
#define  CMD_RUN               37
#define  CMD_ONCHANGE          38
#define  CMD_DOTIMER           39 /* not part of public interface! */
#define  CMD_DURATION          40 /* set and get */


#ifdef HAVE_GLIB

struct XmmsSession 
{
  int session;
  int is_running, is_playing, is_paused, volume;
  int duration, balance, position, length;
  char *filename, *title;  
  GList *callbacks;
} XmmsSession;

struct XmmsEvents 
{
  GList *states;
  int num_callbacks;
  int have_timer;
  int timeout;

} XmmsEvents;

struct XmmsSession *GetSession(int session);
void StartTimer(Tcl_Interp *interp, struct XmmsEvents *ev);
void FreeSession(struct XmmsSession *session);
Tcl_Obj *MakeSessionArray(Tcl_Interp *interp, struct XmmsSession *session);

#else /* ifdef HAVE_GLIB */

struct XmmsEvents 
{
  int timeout;
  int states;
} XmmsEvents;


#endif /* ifdef HAVE_GLIB */

#define EV_PLAYBACK_STATE 1
#define EV_FILE_CHANGE 2
#define EV_POS_CHANGE 3

int IsCommand(Tcl_Interp *interp, Tcl_Obj *CONST obj);

int GetArgIndex(int objc, Tcl_Obj *CONST objv[], char *CONST argname);

int GetIntArg(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int CONST index, int *value);

int Xmms_OnChange(Tcl_Interp *interp, struct XmmsEvents *ev, int objc, Tcl_Obj *CONST objv[],int session);

int Xmms_DoTimer(Tcl_Interp *interp, struct XmmsEvents *ev);

#endif /* XMMSTCL_INT_H */
