/*  XmmsTcl - an xmms remote control library for tcl.
 *
 *  Copyright (C) 2002  Pascal Scheffers <pascal@scheffers.net>
 *
 *    $Id: xmmstcl.c,v 1.4 2002/06/14 07:35:56 pascal Exp $
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "xmmsctrl.h"
#include "config.h"
#include <tcl.h>
#include "xmmstcl.h"
#include "xmmstcl-int.h"
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef HAVE_GLIB
#include <glib.h>
#endif

void XmmsCmdDelete(ClientData clientData) 
{
#ifdef HAVE_GLIB

  struct XmmsEvents *ev=0;
  guint l_iter;

   
  ev=(struct XmmsEvents *)clientData;

  for (l_iter=0; l_iter < g_list_length(ev->states); l_iter++) {
    FreeSession((struct XmmsSession *)g_list_nth(ev->states, l_iter));
  }
#endif
   Tcl_Free((char*)clientData);

}

void XmmsCmdExit(ClientData clientData) 
{
#ifdef HAVE_GLIB

  struct XmmsEvents *ev=0;
  guint l_iter;

   
  ev=(struct XmmsEvents *)clientData;

  for (l_iter=0; l_iter < g_list_length(ev->states); l_iter++) {
    FreeSession((struct XmmsSession *)g_list_nth(ev->states, l_iter));
  }
#endif
   Tcl_Free((char*)clientData);

}

int Xmms_Init(Tcl_Interp *interp) 
{

  struct XmmsEvents *ev=0;

  if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
    return TCL_ERROR;
  }
  
  ev = (struct XmmsEvents *)Tcl_Alloc( sizeof(struct XmmsEvents) );

  ev->states=0;
  ev->num_callbacks=0;
  ev->have_timer=0;
  ev->timeout=MAX_TIMEOUT;
 
  // we should really do some cleanups, but it doesn't seem to work.
  Tcl_CreateObjCommand(interp, "xmms", XmmsCmd, (ClientData*)ev, NULL); //XmmsCmdDelete);
 
  //Tcl_CreateExitHandler(XmmsCmdExit, (ClientData*)ev);

  Tcl_PkgProvide(interp, "xmms", "1.1");

  return TCL_OK;

}


int XmmsCmd(ClientData clientData, 
	    Tcl_Interp *interp,
	    int objc, Tcl_Obj *CONST objv[])
{

  Tcl_Obj *result;
  int session=0, sessionarg=0;
  int cmd;
  result = Tcl_GetObjResult(interp);

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "?command?");
    return TCL_ERROR;
  }

  /* get the session argument, default to 0 */
  session = 0;
  sessionarg = GetArgIndex(objc, objv, "-session");
  if (sessionarg > -1) {
    if (!GetIntArg(interp, objc, objv, sessionarg+1, &session)) {
      result = Tcl_GetObjResult(interp);
      Tcl_SetStringObj(result, E_BAD_SESSION_ARG_MSG, \
		       strlen(E_BAD_SESSION_ARG_MSG));
      return TCL_ERROR;       
    }
  }


  cmd = IsCommand(interp, objv[1]);
  switch(cmd) 
    {
    case CMD_BADCOMMAND:  
      Tcl_SetStringObj(result, E_BAD_CMD_MSG, strlen(E_BAD_CMD_MSG));
      return TCL_ERROR; 
    case CMD_ISRUNNING:
      return Xmms_IsRunning(interp, session); 
    case CMD_RUN:
      return Xmms_Run(interp, objc, objv); 
    case CMD_PLAYLIST:          
      return Xmms_Playlist(interp, objc, objv, session, 0); //replace current
    case CMD_GETVERSION: 
      return Xmms_GetVersion(interp, session);
    case CMD_PLAYLISTADD:        
      return Xmms_Playlist(interp, objc, objv, session, 1); //add to current
    case CMD_PLAYLISTDELETE:     
     return Xmms_PlaylistDelete(interp, objc, objv, session);
    case CMD_PLAY:               
      return Xmms_Play(interp, session);
    case CMD_PAUSE:              
      return Xmms_Pause(interp, session);
    case CMD_STOP:               
      return Xmms_Stop(interp, session);
    case CMD_ISPLAYING:          
       return Xmms_IsPlaying(interp, session);
   case CMD_ISPAUSED:           
      return Xmms_IsPaused(interp, session);
    case CMD_PLAYLISTPOS:         /*set and get */
      return Xmms_PlaylistPos(interp, objc, objv, session);
    case CMD_PLAYLISTLENGTH:     
      return Xmms_Length(interp, session);
    case CMD_PLAYLISTCLEAR:      
      return Xmms_Clear(interp, session);
    case CMD_TIME:                /* set and get */
      return Xmms_Time(interp, objc, objv, session);
    case CMD_DURATION:                /* set and get */
      return Xmms_Duration(interp, objc, objv, session);
    case CMD_VOLUME:              /* set and get */
      return Xmms_Volume(interp, objc, objv, session);
    case CMD_MAINVOLUME:          /* set and get */
      return Xmms_MainVolume(interp, objc, objv, session);
    case CMD_BALANCE:             /* set and get */
      return Xmms_Balance(interp, objc, objv, session);
    case CMD_SKIN:               /* set and get */
      return Xmms_Skin(interp, objc, objv, session);
     case CMD_PLAYLISTFILE:        /* (set and) get */
      return Xmms_File(interp, objc, objv, session);
    case CMD_PLAYLISTTITLE:       /* (set and) get */
      return Xmms_Title(interp, objc, objv, session);
    case CMD_INFO:               /* rate, freq, nchannels */
      return Xmms_Info(interp, session);
    case CMD_MAINWIN:             /* toggle, (set,) get */
      return Xmms_MainWindow(interp, objc, objv, session);
    case CMD_PLWIN:              /* toggle, (set,) get */
      return Xmms_PLWindow(interp, objc, objv, session);
    case CMD_EQWIN:               /* toggle, (set,) get */
      return Xmms_EQWindow(interp, objc, objv, session);
    case CMD_SHOWPREFS:           
      return Xmms_ShowPreferences(interp, session);
    case CMD_AOT:                 /* set */
      return Xmms_AlwaysOnTop(interp, objc, objv, session);
    case CMD_EJECT:             
      return Xmms_Eject(interp, session);
    case CMD_PREV:              
      return Xmms_Previous(interp, session);
    case CMD_NEXT:              
      return Xmms_Next(interp, session);
    case CMD_ADDURL:
      return Xmms_AddUrl(interp, objc, objv, session);             
    case CMD_REPEAT:             /* set, (get) */
      return Xmms_Repeat(interp, objc, objv, session);
    case CMD_SHUFFLE:            /* set, (get) */
      return Xmms_Shuffle(interp, objc, objv, session);
    case CMD_EQ:                  /* (set, get) */
      return Xmms_EQ(interp, objc, objv, session);
    case CMD_EQPREAMP:                  /* (set, get) */
      return Xmms_EQPreamp(interp, objc, objv, session);
    case CMD_EQBAND:                  /* (set, get) */
      return Xmms_EQBand(interp, objc, objv, session);
    case CMD_QUIT:               
      return Xmms_Quit(interp, session);
    case CMD_ONCHANGE:               
      return Xmms_OnChange(interp, clientData, objc, objv, session);
    case CMD_DOTIMER:               
      return Xmms_DoTimer(interp, clientData);
    default:
      Tcl_SetStringObj(result, E_NO_IMPL, strlen(E_NO_IMPL));
      return TCL_ERROR; 
    }

  return TCL_OK;
}


int IsCommand(Tcl_Interp *interp, Tcl_Obj *CONST obj)
{
  char *command;

  if ((command=Tcl_GetString(obj)) == NULL) {
    return CMD_BADCOMMAND;
  }

  if (strcasecmp(command, "isrunning") == 0 ) return CMD_ISRUNNING;
  if (strcasecmp(command, "run") == 0 ) return CMD_RUN;
  if (strcasecmp(command, "playlist") == 0 ) return CMD_PLAYLIST;
  if (strcasecmp(command, "version") == 0 ) return CMD_GETVERSION;
  if (strcasecmp(command, "add") == 0 ) return CMD_PLAYLISTADD;
  if (strcasecmp(command, "delete") == 0 ) return CMD_PLAYLISTDELETE;
  if (strcasecmp(command, "play") == 0 ) return CMD_PLAY;
  if (strcasecmp(command, "pause") == 0 ) return CMD_PAUSE;
  if (strcasecmp(command, "stop") == 0 ) return CMD_STOP;
  if (strcasecmp(command, "isplaying") == 0 ) return CMD_ISPLAYING;
  if (strcasecmp(command, "ispaused") == 0 ) return CMD_ISPAUSED;
  if (strcasecmp(command, "position") == 0 ) return CMD_PLAYLISTPOS;
  if (strcasecmp(command, "length") == 0 ) return CMD_PLAYLISTLENGTH;
  if (strcasecmp(command, "clear") == 0 ) return CMD_PLAYLISTCLEAR;
  if (strcasecmp(command, "duration") == 0 ) return CMD_DURATION;
  if (strcasecmp(command, "time") == 0 ) return CMD_TIME;
  if (strcasecmp(command, "volume") == 0 ) return CMD_VOLUME;
  if (strcasecmp(command, "mainvolume") == 0 ) return CMD_MAINVOLUME;
  if (strcasecmp(command, "balance") == 0 ) return CMD_BALANCE;
  if (strcasecmp(command, "skin") == 0 ) return CMD_SKIN;
  if (strcasecmp(command, "file") == 0 ) return CMD_PLAYLISTFILE;
  if (strcasecmp(command, "title") == 0 ) return CMD_PLAYLISTTITLE;
  if (strcasecmp(command, "info") == 0 ) return CMD_INFO;
  if (strcasecmp(command, "mainwindow") == 0 ) return CMD_MAINWIN;
  if (strcasecmp(command, "plwindow") == 0 ) return CMD_PLWIN;
  if (strcasecmp(command, "eqwindow") == 0 ) return CMD_EQWIN;
  if (strcasecmp(command, "showpreferences") == 0 ) return CMD_SHOWPREFS;
  if (strcasecmp(command, "alwaysontop") == 0 ) return CMD_AOT;
  if (strcasecmp(command, "eject") == 0 ) return CMD_EJECT;
  if (strcasecmp(command, "previous") == 0 ) return CMD_PREV;
  if (strcasecmp(command, "next") == 0 ) return CMD_NEXT;
  if (strcasecmp(command, "addurl") == 0 ) return CMD_ADDURL;
  if (strcasecmp(command, "repeat") == 0 ) return CMD_REPEAT;
  if (strcasecmp(command, "shuffle") == 0 ) return CMD_SHUFFLE;
  if (strcasecmp(command, "eq") == 0 ) return CMD_EQ;
  if (strcasecmp(command, "eqpreamp") == 0 ) return CMD_EQPREAMP;
  if (strcasecmp(command, "eqband") == 0 ) return CMD_EQBAND;
  if (strcasecmp(command, "quit") == 0 ) return CMD_QUIT;
  if (strcasecmp(command, "onchange") == 0 ) return CMD_ONCHANGE;
  if (strcasecmp(command, "dotimer") == 0 ) return CMD_DOTIMER;

  return CMD_BADCOMMAND;
}


int Xmms_IsRunning(Tcl_Interp *interp, 
		   int session)
{

  Tcl_Obj *result;
  
  result = Tcl_GetObjResult(interp);

  Tcl_SetIntObj(result, xmms_remote_is_running( session ));

  return TCL_OK;
}

int Xmms_Run(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
  /*
   * a new xmms session just loops over session ids starting at 0
   * to find the first one that is not running and it then uses that
   * (see xmmssources/xmms/controlsocket.c).
   * We'll do the same to get the expected new session id, start xmms
   * and hope for the best...
   */
  int exp_session=0, i, patharg_ix, timeout_ix;
  Tcl_Obj *result;
  pid_t new_proc=0;
  int timeout=MAX_EXEC_WAIT_TIME;
  char *xmms_cmd=0, **xmms_argv=0;

  result = Tcl_GetObjResult(interp);

  for (exp_session=0; exp_session < MAX_EXPECTED_SESSIONS; exp_session++) {
    if ( !xmms_remote_is_running(exp_session) ) break;
  }

  if ( exp_session >= MAX_EXPECTED_SESSIONS ) {    
   Tcl_SetStringObj(result, E_NO_MORE_SESSION_MSG, strlen(E_NO_MORE_SESSION_MSG));
   return TCL_ERROR;
  }

  patharg_ix = 2; //set the default patharg_ix "xmms run PATH ..."
  // timeout specified?
  if ((timeout_ix=GetArgIndex(objc, objv, "-timeout")) != -1) {
    patharg_ix=timeout_ix+2; // 'xmms run -timeout 1234 PATH ...'
    if (!GetIntArg(interp, objc, objv, timeout_ix+1, &timeout)==-1) {
      Tcl_SetStringObj(result, E_BAD_TIMEOUT_ARG_MSG, \
		       strlen(E_BAD_TIMEOUT_ARG_MSG));
      return TCL_ERROR;      
    }
  } 

  // path and args specified?
  if (patharg_ix < objc) {

    xmms_cmd = Tcl_GetString(objv[patharg_ix]);
    xmms_argv = (char **)Tcl_Alloc(sizeof(char*) * (objc-patharg_ix+1) );

    for (i=0; i < (objc-patharg_ix); i++) {      
      xmms_argv[i]=Tcl_GetString(objv[patharg_ix+i]);
    }    
    xmms_argv[objc-patharg_ix]=NULL;

  } else {
    //set default command array.
    xmms_cmd = "xmms";
    xmms_argv = (char **)Tcl_Alloc(sizeof(char*) * 2);
    xmms_argv[0] = xmms_cmd;
    xmms_argv[1] = NULL;
  }


  new_proc = fork();
  if (new_proc == 0) {
    //child process
    execvp(xmms_cmd, xmms_argv);
    // just make sure.
    exit(0);
  }

  if (xmms_argv) Tcl_Free((char *)xmms_argv);

  if (new_proc == -1) {
    Tcl_SetStringObj(result, E_FORK_FAILED_MSG, strlen(E_FORK_FAILED_MSG));
    return TCL_ERROR;
    //fork failed.
  }

  //now, while the pid is running and for a maximum of FORK_TIMEOUT seconds
  //see if a new session comes up:
  i=0;
  while (i * 50 < timeout) {
    i++;
    if ( xmms_remote_is_running(exp_session) ) break;

    Tcl_Sleep(50); // milliseconds.
  }  

  if ( !xmms_remote_is_running(exp_session) && timeout > 0) {    
   Tcl_SetStringObj(result, E_EXEC_TIMEOUT_MSG, strlen(E_EXEC_TIMEOUT_MSG));
   return TCL_ERROR;
  }

  result = Tcl_GetObjResult(interp);

  Tcl_SetIntObj(result, exp_session);

  return TCL_OK;
}


int Xmms_GetVersion(Tcl_Interp *interp, 
		   int session)
{

  Tcl_Obj *result;
  
  result = Tcl_GetObjResult(interp);

  Tcl_SetIntObj(result, xmms_remote_get_version( session ));

  return TCL_OK;
}

int Xmms_Play(Tcl_Interp *interp, 
		   int session)
{

  Tcl_Obj *result;
  
  result = Tcl_GetObjResult(interp);
  
  xmms_remote_play( session );

  Tcl_SetIntObj(result, xmms_remote_is_playing( session ));

  return TCL_OK;
}

int Xmms_Next(Tcl_Interp *interp, 
		   int session)
{

 
  xmms_remote_playlist_next( session );

  return TCL_OK;
}

int Xmms_Previous(Tcl_Interp *interp, 
		   int session)
{

  xmms_remote_playlist_prev( session );

  return TCL_OK;
}

int Xmms_Eject(Tcl_Interp *interp, 
		   int session)
{
  xmms_remote_eject( session );

  return TCL_OK;
}

int Xmms_Length(Tcl_Interp *interp, 
		   int session)
{

  Tcl_Obj *result;
  
  result = Tcl_GetObjResult(interp);  

  Tcl_SetIntObj(result, xmms_remote_get_playlist_length( session ));

  return TCL_OK;
}

int Xmms_Clear(Tcl_Interp *interp, 
		   int session)
{

  xmms_remote_playlist_clear( session );

  return TCL_OK;
}
int Xmms_Quit(Tcl_Interp *interp, 
		   int session)
{

  xmms_remote_quit( session );

  return TCL_OK;
}

int Xmms_Pause(Tcl_Interp *interp, 
		   int session)
{

  Tcl_Obj *result;
  
  result = Tcl_GetObjResult(interp);
  
  xmms_remote_pause( session );

  Tcl_SetIntObj(result, !xmms_remote_is_playing( session ));

  return TCL_OK;
}

int Xmms_Stop(Tcl_Interp *interp, 
		   int session)
{

  Tcl_Obj *result;
  
  result = Tcl_GetObjResult(interp);
  
  xmms_remote_stop( session );

  Tcl_SetIntObj(result, !xmms_remote_is_playing( session ));

  return TCL_OK;
}

int Xmms_IsPlaying(Tcl_Interp *interp, 
		   int session)
{

  Tcl_Obj *result;
  
  result = Tcl_GetObjResult(interp);
  
  Tcl_SetIntObj(result, xmms_remote_is_playing( session ));

  return TCL_OK;
}

int Xmms_IsPaused(Tcl_Interp *interp, 
		   int session)
{

  Tcl_Obj *result;
  
  result = Tcl_GetObjResult(interp);
  
  Tcl_SetIntObj(result, xmms_remote_is_paused( session ));

  return TCL_OK;
}

int Xmms_Playlist(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], 
		  int session, int enqueue)
     /* 
      * sets (and gets) the playlist  
      */     
{
  int i;
  char *test=0, *tmp=0;
  int len, gettitles=0;  
  gchar **list=0;

  int l_objc=-1;
  Tcl_Obj **l_objv;

  Tcl_Obj **r_objv=0;

  if (objc >= 3) {    
    if ((test = Tcl_GetStringFromObj(objv[2], &len))!=NULL) {
      // Test to see if the second argument doesn't start with a '-'
      // this would indicate something like '-session'
      if (test[0] != '-') {

	if (Tcl_ListObjGetElements(interp, objv[2], &l_objc, &l_objv)==TCL_ERROR) {
	  return TCL_ERROR;
	}

	//build an array of pointers to the strings
	list = (gchar **)Tcl_Alloc(sizeof(gchar*) * l_objc);
	for (i=0; i < l_objc; i++ ) {
	  list[i] = Tcl_GetStringFromObj(l_objv[i], &len);	  
	}
      }
    }
  }

  // Do we set/append to the playlist? only if there are list elements.
  if (l_objc > 0) {
    xmms_remote_playlist(session, list, l_objc, enqueue);
    Tcl_Free((char*)list);
  } 

  gettitles=GetArgIndex(objc, objv, "-titles");
  len = xmms_remote_get_playlist_length(session);

  if (len > 0) {
    r_objv = (Tcl_Obj **) Tcl_Alloc(len * sizeof(Tcl_Obj*));

    for (i=0; i<len; i++) {
      if (gettitles != -1) {
	tmp = xmms_remote_get_playlist_title(session, i);
      } else {
	tmp = xmms_remote_get_playlist_file(session, i);
      }
      r_objv[i]=Tcl_NewStringObj(tmp, strlen(tmp)); 
      if (tmp) g_free(tmp);
    }


    Tcl_SetObjResult(interp, Tcl_NewListObj(len, r_objv));  

    if (r_objv) Tcl_Free((char *)r_objv);
  }

  return TCL_OK;

}

int Xmms_PlaylistPos(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],
		     int session)
     /* 
      * sets and gets the playlist position 
      * caller can check if it worked by comparing the returned position 
      * to the specified position.
      */     
{
  int newpos=-1;

  Tcl_Obj *result;

  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &newpos)==TCL_ERROR) {
      newpos=-1;
    }
  }

  if (newpos > -1) {
    // try to set
    xmms_remote_set_playlist_pos(session, newpos);
  } 
 
  result = Tcl_GetObjResult(interp);
  
  Tcl_SetIntObj(result, xmms_remote_get_playlist_pos( session ));  

  return TCL_OK;
}

int Xmms_PlaylistDelete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],
			int session)
     /* 
      * deletes an item from the playlist, defaults to current
      */     
{
  int pos=-1;


  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &pos)==TCL_ERROR) {
      pos=-1;
    }
  }

  if (pos == -1) {
    pos = xmms_remote_get_playlist_pos(session);
  } 

  xmms_remote_playlist_delete(session, pos);

  return TCL_OK;
}


int Xmms_File(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * gets the current file(default), or the file at '-pos n'
      * sets the player to a file '-set filename' (TODO)
      * a set that fails will raise an error.
      */     
{
  char *file;
  int posarg=-1, pos=-1;
  Tcl_Obj *result;

  posarg = GetArgIndex(objc, objv, "-pos");
  if (posarg > -1) {
    if (!GetIntArg(interp, objc, objv, posarg+1, &pos)) {
      result = Tcl_GetObjResult(interp);
      Tcl_SetStringObj(result, E_BAD_POS_ARG_MSG, strlen(E_BAD_POS_ARG_MSG));
      return TCL_ERROR;       
    }
  }

  if (pos < 0 ) {
    pos = xmms_remote_get_playlist_pos(session);
  }


  result = Tcl_GetObjResult(interp);
  
  file = xmms_remote_get_playlist_file( session, pos );
  if (file) {
    Tcl_SetStringObj(result, file, strlen(file));  
    g_free(file);
  }

  return TCL_OK;
}

int Xmms_Title(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session)
     /* 
      * gets the current tile(default), or the title at '-pos n'
      * sets the player to a title '-set title' (TODO)
      * a set that fails will raise an error.
      */     
{
  char *title;
  int posarg=-1, pos=-1;
  Tcl_Obj *result;

  posarg = GetArgIndex(objc, objv, "-pos");
  if (posarg > -1) {
    if (!GetIntArg(interp, objc, objv, posarg+1, &pos)) {
      result = Tcl_GetObjResult(interp);
      Tcl_SetStringObj(result, E_BAD_POS_ARG_MSG, strlen(E_BAD_POS_ARG_MSG));
      return TCL_ERROR;       
    }
  }

  if (pos < 0 ) {
    pos = xmms_remote_get_playlist_pos(session);
  }


  result = Tcl_GetObjResult(interp);
  
  title = xmms_remote_get_playlist_title( session, pos );
  if (title) {
    Tcl_SetStringObj(result, title, strlen(title));  
    g_free(title);
  }

  return TCL_OK;
}

int Xmms_Time(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * sets and gets the playing time index
      * caller can check if it worked by comparing the returned position 
      * to the specified position.
      */     
{
  int newpos=-1;

  Tcl_Obj *result;

  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &newpos)==TCL_ERROR) {
      newpos=-1;
    }
  }

  if (newpos > -1) {
    // try to set
    xmms_remote_jump_to_time(session, newpos);
  } 
 
  result = Tcl_GetObjResult(interp);
  
  Tcl_SetIntObj(result, xmms_remote_get_output_time( session ));  

  return TCL_OK;
}

int Xmms_Duration(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
{
  int pos=-1;

  Tcl_Obj *result;

  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &pos)==TCL_ERROR) {
      pos=-1;
    }
  }
  
  if (pos==-1) pos=xmms_remote_get_playlist_pos(session);

  result = Tcl_GetObjResult(interp);
  
  Tcl_SetIntObj(result, xmms_remote_get_playlist_time( session, pos ));  

  return TCL_OK;
}


int Xmms_Volume(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * sets and gets the session volume
      * caller can check if it worked by comparing the returned position 
      * to the specified position.
      */     
{
  int vl=-1;
  int vr=-1;

  Tcl_Obj *vol[2];

  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &vl)==TCL_ERROR) {
      vl=-1;
    }
  }

  vr=vl;

  if (objc >= 4 && vl > -1 ) {
    if (Tcl_GetIntFromObj(interp, objv[3], &vr)==TCL_ERROR) {
      vr=vl;
    }
  }

  if (vl > -1) {
    // try to set
    xmms_remote_set_volume(session, vl, vr);
  } 

  xmms_remote_get_volume(session, &vl, &vr);

  vol[0] = Tcl_NewIntObj(vl);
  vol[1] = Tcl_NewIntObj(vr);

  Tcl_SetObjResult(interp, Tcl_NewListObj(2, vol));
  

  return TCL_OK;
}

int Xmms_MainVolume(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],
		    int session)
     /* 
      * sets and gets the session volume
      * caller can check if it worked by comparing the returned position 
      * to the specified position.
      */     
{
  int vl=-1;

  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &vl)==TCL_ERROR) {
      vl=-1;
    }
  }
 
  if (vl > -1) {
    // try to set
    xmms_remote_set_main_volume(session, vl);
  } 

  vl = xmms_remote_get_main_volume(session);

  Tcl_SetObjResult(interp, Tcl_NewIntObj(vl));
 

  return TCL_OK;
}

int Xmms_Balance(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * sets and gets the session balance
      * caller can check if it worked by comparing the returned position 
      * to the specified position.
      */     
{
  int balance=-1000;

  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &balance)==TCL_ERROR) {
      balance=-1000;
    }
  }
 
  if (balance > -1000) {
    // try to set
    xmms_remote_set_balance(session, balance);
  } 

  balance = xmms_remote_get_balance(session);

  Tcl_SetObjResult(interp, Tcl_NewIntObj(balance));
 

  return TCL_OK;
}

int Xmms_EQ(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      */     
{
  char *test=0;
  int len, i;
  Tcl_Obj *rlist[11];
  double dtmp;

  int l_objc=-1;
  Tcl_Obj **l_objv;


  if (objc >= 3) {    
    if ((test = Tcl_GetStringFromObj(objv[2], &len))!=NULL) {
      // Test to see if the second argument doesn't start with a '-'
      // this would indicate something like '-session' '-flat'
      if (test[0] != '-') {

	if (Tcl_ListObjGetElements(interp, objv[2], &l_objc, &l_objv)==TCL_ERROR) {
	  return TCL_ERROR;
	}
	if (l_objc != 11) {
	  Tcl_SetObjResult(interp, Tcl_NewStringObj(E_BAD_EQ_LIST_MSG, 
						    strlen(E_BAD_EQ_LIST_MSG)));
	}

	//build an array of pointers to the strings
	//list = (gfloat **)Tcl_Alloc(sizeof(gfloat) * l_objc);
	for (i=1; i < l_objc; i++ ) {
	  //gfloat is not a double, so we need a double-tempvar
	  if (Tcl_GetDoubleFromObj(interp, l_objv[i], &dtmp)==TCL_ERROR) return TCL_ERROR;    
	  xmms_remote_set_eq_band(session, i-1, (gfloat)dtmp);
	} 
	if (Tcl_GetDoubleFromObj(interp, l_objv[0], &dtmp)==TCL_ERROR) return TCL_ERROR; 
	xmms_remote_set_eq_preamp(session, (gfloat)dtmp);
      } else {

	i = GetArgIndex(objc, objv, "-flat");
	if ( i > 0 ) {
	  dtmp=0;
	  if (i+1 < objc) {
	    if (Tcl_GetDoubleFromObj(interp, objv[i+1], &dtmp)==TCL_ERROR) dtmp=0;

	  }
	  for(i=0; i<10; i++) xmms_remote_set_eq_band(session, i, dtmp);
	  xmms_remote_set_eq_preamp(session, dtmp);
	}
      }
    }
  }
 

  //should be possible to do this with:
  //xmms_remote_get_eq(session, &preamp, &bands);
  //but a loop is simpler, not very expensive, ets. 
  //so, lets just do a loop over xmms_remote_get_eq_preamp and 
  //xmms_remote_get_eq_band:

  rlist[0] = Tcl_NewDoubleObj((double)xmms_remote_get_eq_preamp(session));
  for (i=0; i<10; i++) {
    rlist[i+1] = Tcl_NewDoubleObj((double)xmms_remote_get_eq_band(session, i));    
  }

  Tcl_SetObjResult(interp, Tcl_NewListObj(11, rlist));

  return TCL_OK;
}

int Xmms_EQPreamp(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * sets and gets the session volume
      * caller can check if it worked by comparing the returned position 
      * to the specified position.
      */     
{
  double preamp=-100;

  if (objc >= 3) {
    if (Tcl_GetDoubleFromObj(interp, objv[2], &preamp)==TCL_ERROR) {
      preamp=-100;
    }
  }
 
  if (preamp > -100) {
    // try to set
    xmms_remote_set_eq_preamp(session, preamp);
  } 

  preamp = xmms_remote_get_eq_preamp(session);

  Tcl_SetObjResult(interp, Tcl_NewDoubleObj(preamp));
 

  return TCL_OK;
}

int Xmms_EQBand(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * sets and gets the session volume
      * caller can check if it worked by comparing the returned position 
      * to the specified position.
      */     
{
  int band=-1;
  double level=-100;

  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &band)==TCL_ERROR) {
      band=-100;
    }
  }

  if (band < 0 || band > 9) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj(E_BAD_EQ_BAND_MSG, 
					      strlen(E_BAD_EQ_BAND_MSG)));
    return TCL_ERROR;
  }

  if (objc >= 4) {
    if (Tcl_GetDoubleFromObj(interp, objv[3], &level)==TCL_ERROR) {
      level=-100;
    }
  }

  
 
  if (level > -100) {
    // try to set
    xmms_remote_set_eq_band(session, band, level);
  } 

  level = xmms_remote_get_eq_band(session, band);

  Tcl_SetObjResult(interp, Tcl_NewDoubleObj(level));
 

  return TCL_OK;
}

int Xmms_Info(Tcl_Interp *interp, int session)
     /* 
      * gets bitrate, sample freq. and number of channels
      * returns a list in that order.
      */     
{

  Tcl_Obj *info[3];
  int rate, freq, nch;

  xmms_remote_get_info(session, &rate, &freq, &nch);

  info[0] = Tcl_NewIntObj(rate);
  info[1] = Tcl_NewIntObj(freq);
  info[2] = Tcl_NewIntObj(nch);

  Tcl_SetObjResult(interp, Tcl_NewListObj(3, info));
  

  return TCL_OK;
}

int Xmms_Repeat(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * get and sets the repeat state
      */     
{
  int newstate=-1;

  Tcl_Obj *result;
  int cur;


  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &newstate)==TCL_ERROR) {
      newstate=-1;
    }
  }

  cur = xmms_remote_is_repeat( session );

  if (newstate > -1) {
    // try to set
    if ( newstate == 1 && cur == 0 ) xmms_remote_toggle_repeat(session);
    if ( newstate == 0 && cur == 1 ) xmms_remote_toggle_repeat(session);   

  } 
 
  result = Tcl_GetObjResult(interp);
  
  Tcl_SetIntObj(result, xmms_remote_is_repeat( session ));  

  return TCL_OK;
}


int Xmms_Shuffle(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * sets and gets shuffle state
      */     
{
  int newstate=-1;

  Tcl_Obj *result;
  int cur;


  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &newstate)==TCL_ERROR) {
      
      newstate=-1;
    }
  }

  cur = xmms_remote_is_shuffle( session );

    if (newstate > -1) {
      // try to set
      if ( newstate == 1 && cur == 0 ) xmms_remote_toggle_shuffle(session);
      if ( newstate == 0 && cur == 1 ) xmms_remote_toggle_shuffle(session); 

    } 
 
  result = Tcl_GetObjResult(interp);
  
  Tcl_SetIntObj(result, xmms_remote_is_shuffle( session ));  

  return TCL_OK;
}

int Xmms_MainWindow(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * sets and gets main window visibility state
      */     
{
  int newstate=-1;

  Tcl_Obj *result;
  int cur;


  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &newstate)==TCL_ERROR) {
      
      newstate=-1;
    }
  }

  cur = xmms_remote_is_main_win( session );

    if (newstate > -1) {
      // try to set
      if ( newstate == 1 && cur == 0 ) xmms_remote_main_win_toggle(session, newstate);
      if ( newstate == 0 && cur == 1 ) xmms_remote_main_win_toggle(session, newstate); 

    } 
 
  result = Tcl_GetObjResult(interp);
  
  Tcl_SetIntObj(result, xmms_remote_is_main_win( session ));  

  return TCL_OK;
}

int Xmms_PLWindow(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * sets and gets playlist window visibility state
      */     
{
  int newstate=-1;

  Tcl_Obj *result;
  int cur;


  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &newstate)==TCL_ERROR) {
      
      newstate=-1;
    }
  }

  cur = xmms_remote_is_pl_win( session );

    if (newstate > -1) {
      // try to set
      if ( newstate == 1 && cur == 0 ) xmms_remote_pl_win_toggle(session, newstate);
      if ( newstate == 0 && cur == 1 ) xmms_remote_pl_win_toggle(session, newstate); 

    } 
 
  result = Tcl_GetObjResult(interp);
  
  Tcl_SetIntObj(result, xmms_remote_is_pl_win( session ));  

  return TCL_OK;
}

int Xmms_EQWindow(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * sets and gets main window visibility state
      */     
{
  int newstate=-1;

  Tcl_Obj *result;
  int cur;


  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &newstate)==TCL_ERROR) {
      
      newstate=-1;
    }
  }

  cur = xmms_remote_is_eq_win( session );

    if (newstate > -1) {
      // try to set
      if ( newstate == 1 && cur == 0 ) xmms_remote_eq_win_toggle(session, newstate);
      if ( newstate == 0 && cur == 1 ) xmms_remote_eq_win_toggle(session, newstate); 

    } 
 
  result = Tcl_GetObjResult(interp);
  
  Tcl_SetIntObj(result, xmms_remote_is_eq_win( session ));  

  return TCL_OK;
}

int Xmms_ShowPreferences(Tcl_Interp *interp, 
		   int session)
{

  xmms_remote_show_prefs_box( session );

  return TCL_OK;
}

int Xmms_AlwaysOnTop(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      * sets the always on top state 
      */     
{
  int newstate=-1;

  if (objc >= 3) {
    if (Tcl_GetIntFromObj(interp, objv[2], &newstate)==TCL_ERROR) return TCL_ERROR;
  }
  
  if (newstate == 0 || newstate == 1)  xmms_remote_toggle_aot(session, newstate);

  return TCL_OK;
}

int Xmms_AddUrl(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session)
     /* 
      */     
{
  gchar *url=NULL;
  int len;

  if (objc >= 3) {
    if ((url=Tcl_GetStringFromObj(objv[2], &len))==NULL) {
      return TCL_ERROR;
    }
  }
 
  if (url != NULL) {
    xmms_remote_playlist_add_url_string(session, url);
  } 

  return TCL_OK;
}

int Xmms_Skin(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session)
     /* 
      * gets the current tile(default), or the title at '-pos n'
      * sets the player to a title '-set title' (TODO)
      * a set that fails will raise an error.
      */     
{
  char *skin=0;
  Tcl_Obj *result;
  int len;

  if (objc >= 3) {
    if ((skin=Tcl_GetStringFromObj(objv[2], &len))==NULL) {
      return TCL_ERROR;
    }
    //is it just the '-session' arg?
    if (skin[0] == '-') skin=0;

  }

  if (skin) {
    xmms_remote_set_skin(session, skin);
  }

  result = Tcl_GetObjResult(interp);
  
  skin = xmms_remote_get_skin( session );
  if (skin) Tcl_SetStringObj(result, skin, strlen(skin));  

  return TCL_OK;
}

int Xmms_OnChange(Tcl_Interp *interp, struct XmmsEvents *ev, int objc, Tcl_Obj *CONST objv[], int session)
{
#ifdef HAVE_GLIB

  char *callback_proc=0, *tmp=0;  
  GList *l_item=0;
  int len, do_cancel=0;
  guint l_iter;

  struct XmmsSession *ThisSession=NULL;

  if (objc >= 3) {
    //callback_proc=Tcl_DuplicateObj(objv[2]);
    if ((callback_proc=Tcl_GetStringFromObj(objv[2], &len))==NULL) {
      return TCL_ERROR;
    }
    //is it just the '-session' or '-cancel' arg?
    if (strcasecmp(callback_proc, "-session") == 0 ) callback_proc=0;
    else if (strcasecmp(callback_proc, "-cancel") == 0 ) callback_proc=0;
  }

  if (!callback_proc) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj(E_ONCHANGE_USAGE_MSG, \
					      strlen(E_ONCHANGE_USAGE_MSG))); 
    return TCL_ERROR;
  }

  //cancel?
  do_cancel=( GetArgIndex(objc, objv, "-cancel") > -1 );

  //do we have this session in the list?
  for (l_iter=0; l_iter < g_list_length(ev->states); l_iter++) {
    l_item = g_list_nth(ev->states, l_iter);
    ThisSession = (struct XmmsSession *)l_item->data;
    if (ThisSession->session == session) break; // yes found it! break.
    ThisSession=NULL; //nope, this isn't it - set null again. 
  }
  
  if (!ThisSession && !do_cancel) {
    //get the new session and add it to the list.
    ThisSession = GetSession(session); 
    if (!ThisSession) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj(E_NO_MEM_MSG, \
					      strlen(E_NO_MEM_MSG))); 
      return TCL_ERROR;

    }
    ev->states=g_list_append(ev->states, ThisSession);
  }

  if (ThisSession && !do_cancel) {
    //We have a session and we don't want to cancel.

    //let's not add duplicates.
    for (l_iter=0; l_iter < g_list_length(ThisSession->callbacks); l_iter++) {
      l_item = g_list_nth(ThisSession->callbacks, l_iter);
      tmp = (char *)l_item->data;
      if (strcasecmp(tmp, callback_proc) == 0) return TCL_OK; // duplicate, silently fail.
      tmp=NULL; //nope, this isn't it - set null again. 
    }   

    //Lets add the callback proc to the list.
    callback_proc=strdup(callback_proc);
    if (callback_proc) {
      //debug_print("onchange added: ");
      //debug_print(callback_proc);
      //debug_print("\n");

      ThisSession->callbacks=g_list_append(ThisSession->callbacks, callback_proc);
  
      //And increase the reference count.
      ev->num_callbacks++;

      if (!ev->have_timer) StartTimer(interp, ev);
      //else debug_print("Have TIMER!\n");

    } else { 
      //oh dear, strdup failed.
      Tcl_SetObjResult(interp, Tcl_NewStringObj(E_NO_MEM_MSG, \
						strlen(E_NO_MEM_MSG))); 
      return TCL_ERROR;
    }
  } else if (ThisSession && do_cancel) {
    //remove the callback

    //first, find it in the list:
    for (l_iter=0; l_iter < g_list_length(ThisSession->callbacks); l_iter++) {
      l_item = g_list_nth(ThisSession->callbacks, l_iter);
      tmp = (char *)l_item->data;
      if (strcasecmp(tmp, callback_proc) == 0) break; // yes found it! break.
      tmp=NULL; //nope, this isn't it - set null again. 
    }
    if (tmp) {
      //debug_print("onchange drop: ");
      //debug_print(tmp);
      //debug_print("\n");
      ThisSession->callbacks=g_list_remove(ThisSession->callbacks, tmp);
      ev->num_callbacks--;    
    } 
    

  }
  return TCL_OK;

#else 
  Tcl_SetObjResult(interp, Tcl_NewStringObj(E_NO_IMPL, strlen(E_NO_IMPL))); 
  return TCL_ERROR;
#endif
}

int GetArgIndex(int objc, Tcl_Obj *CONST objv[], char *CONST argname)
     /* 
      * Return objc index for argname, -1 if argument does not exist
      * (start at index 1)
      */
{
  int i;
  char *thisarg;

  for (i=1; i<objc; i++) {
    if ((thisarg=Tcl_GetString(objv[i])) != NULL) {
      if ( strcasecmp(thisarg, argname) == 0 ) {
	return i;
      }
    }    
  }
  return -1;
}


int GetIntArg(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int CONST index, int *value)
{
  if ( index > objc || index < 0) {
    return 0;
  }

  if (Tcl_GetIntFromObj(interp, objv[index], value)==TCL_ERROR) {
    return 0;
  }  
  return 1;
}

#ifdef HAVE_GLIB

struct XmmsSession *GetSession(int session) 
{
  struct XmmsSession *newSession=NULL;
  
  if (!(newSession=(struct XmmsSession *)Tcl_Alloc(sizeof(struct XmmsSession)))) return NULL;

  newSession->session = session;
  if ( xmms_remote_is_running( session ) ) {
    newSession->is_running=1;
    newSession->is_playing=xmms_remote_is_playing(session);
    newSession->is_paused=xmms_remote_is_paused(session);
    newSession->position=xmms_remote_get_playlist_pos(session);
    newSession->volume=xmms_remote_get_main_volume(session);
    newSession->balance=xmms_remote_get_balance(session);
    newSession->duration=xmms_remote_get_playlist_time(session, newSession->position);
    newSession->length=xmms_remote_get_playlist_length(session);
    newSession->filename=xmms_remote_get_playlist_file(session, newSession->position);
    newSession->title=xmms_remote_get_playlist_title(session, newSession->position);
    newSession->callbacks=NULL;
  } else {
    newSession->is_running=0;
    newSession->is_playing=0;
    newSession->is_paused=0;
    newSession->position=-1;
    newSession->duration=-1;
    newSession->length=-1;
    newSession->volume=0;
    newSession->balance=0;
    newSession->filename=NULL;
    newSession->title=NULL;
    newSession->callbacks=NULL;
  }


  return newSession;
  
}

void FreeSession(struct XmmsSession *session) {
  
  GList *l_item=0;
  guint l_iter;
  
  if (!session) return; // yeah, right.


  //loop through all callbacks and free them
  for (l_iter=0; l_iter < g_list_length(session->callbacks); l_iter++) {
    l_item = g_list_nth(session->callbacks, l_iter);    
    Tcl_Free((char*)l_item->data);      
  }
 
 //free the g_list:
  if (session->callbacks) g_list_free(session->callbacks);

  //free the (possible) filename and title buffers.
  if (session->filename) g_free(session->filename);
  if (session->title) g_free(session->title);

  //and free the session struct itself:
  free(session);

}

Tcl_Obj *MakeSessionArray(Tcl_Interp *interp, struct XmmsSession *session)
{
  //not an actual tcl arrar but an even numbered tcl list, 
  //with 'name value name value', in the interp: 
  //just 'array set arrayName <list>'
  
  Tcl_Obj *list;

  list=Tcl_NewListObj(0, NULL); //empty list.
  
  //name: session
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewStringObj("session", strlen("session") ) )==TCL_ERROR ) return NULL;
  //value:
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewIntObj(session->session) )==TCL_ERROR ) return NULL;

  //name: isrunning
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewStringObj("isrunning", strlen("isrunning") ) )==TCL_ERROR ) return NULL;
  //value:
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewIntObj(session->is_running) )==TCL_ERROR ) return NULL;

  //name: isplaying
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewStringObj("isplaying", strlen("isplaying") ) ) ==TCL_ERROR ) return NULL;
  //value:
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewIntObj(session->is_playing) ) ==TCL_ERROR ) return NULL;

  //name: 
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewStringObj("ispaused", strlen("ispaused") ) ) ==TCL_ERROR ) return NULL;
  //value:
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewIntObj(session->is_paused) ) ==TCL_ERROR ) return NULL;

  //name: 
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewStringObj("duration", strlen("duration") ) ) ==TCL_ERROR ) return NULL;
  //value:
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewIntObj(session->duration) ) ==TCL_ERROR ) return NULL;
  //name: 
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewStringObj("position", strlen("position") ) ) ==TCL_ERROR ) return NULL;
  //value:
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewIntObj(session->position) ) ==TCL_ERROR ) return NULL;
  //name: 
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewStringObj("length", strlen("length") ) ) ==TCL_ERROR ) return NULL;
  //value:
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewIntObj(session->length) ) ==TCL_ERROR ) return NULL;

  //name: 
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewStringObj("volume", strlen("volume") ) ) ==TCL_ERROR ) return NULL;
  //value:
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewIntObj(session->volume) ) ==TCL_ERROR ) return NULL;

  //name: 
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewStringObj("balance", strlen("balance") ) ) ==TCL_ERROR ) return NULL;
  //value:
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewIntObj(session->balance) ) ==TCL_ERROR ) return NULL;


  //name: file
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewStringObj("file", strlen("file") ) ) ==TCL_ERROR ) return NULL;
  if (session->filename) {
    //value:
    if (Tcl_ListObjAppendElement(interp, list, \
				 Tcl_NewStringObj(session->filename, strlen(session->filename) ) ) ==TCL_ERROR ) return NULL;
  } else {
    //value: empty string
    if (Tcl_ListObjAppendElement(interp, list, \
				 Tcl_NewStringObj("", strlen("") ) ) ==TCL_ERROR ) return NULL;
  }



  //name: 
  if (Tcl_ListObjAppendElement(interp, list, \
			       Tcl_NewStringObj("title", strlen("title") ) ) ==TCL_ERROR ) return NULL;
  if (session->title) {
    //value:
    if (Tcl_ListObjAppendElement(interp, list, \
				 Tcl_NewStringObj(session->title, strlen(session->title) ) ) ==TCL_ERROR ) return NULL;
  } else {
    //value: empty string
    if (Tcl_ListObjAppendElement(interp, list, \
				 Tcl_NewStringObj("", strlen("") ) ) ==TCL_ERROR ) return NULL;
  }


  return list;

}


void StartTimer(Tcl_Interp *interp, struct XmmsEvents *ev) {

  char aftercmd[200]; //better too large than overflow...

  sprintf(aftercmd, "after %d xmms dotimer", ev->timeout);
  //debug_print(aftercmd);

  ev->have_timer=1;

  Tcl_Eval(interp, aftercmd);

  ev->timeout += TIMEOUT_INCR;
  if (ev->timeout > MAX_TIMEOUT) ev->timeout=MAX_TIMEOUT;

  return;
}
#endif // define HAVE_GLIB

int Xmms_DoTimer(Tcl_Interp *interp, struct XmmsEvents *ev) 
{
  
  GList *l_item=0, *cb_item=0;
  guint l_iter, cb_iter;
  Tcl_Obj *arr_obj, *cmd_objv[4];
  struct XmmsSession *ThisSession, *newSession;
  int changes=0;
  ev->have_timer=0;
  //debug_print("\nDo timer\n");
  for (l_iter=0; l_iter < g_list_length(ev->states); l_iter++) {
    l_item = g_list_nth(ev->states, l_iter);
    ThisSession=(struct XmmsSession *)l_item->data;
    //should we even bother?
    if (g_list_length(ThisSession->callbacks) > 0) {
      //yes, there are callbacks.
      newSession=GetSession(ThisSession->session);
      if (newSession) {
	changes=0;
	if (newSession->is_running != ThisSession->is_running) changes++;
	if (newSession->is_playing != ThisSession->is_playing) changes++;
	if (newSession->is_paused != ThisSession->is_paused) changes++;
	if (newSession->position != ThisSession->position ) changes++;
	if (newSession->duration != ThisSession->duration) changes++;
	if (newSession->length != ThisSession->length) changes++;
	if (newSession->volume != ThisSession->volume) changes++;
	if (newSession->balance != ThisSession->balance) changes++;
	
	//if one of them is null, it's a change.
	if ( (newSession->filename==0) != (ThisSession->filename==0) ) changes++;
	if ( (newSession->title==0) != (ThisSession->title==0) ) changes++;
	
	//do string compares:
	if (!changes) { // don't bother to do strcmp if we already have changes.
	  if ( newSession->filename && ThisSession->filename ) {
	    if (strcmp(newSession->filename, ThisSession->filename) != 0 ) changes++;
	  }
	  if ( newSession->title && ThisSession->title ) {
	    if (strcmp(newSession->title, ThisSession->title) != 0 ) changes++;  
	  } 
	}
	
	if (changes) {
	  ev->timeout=MIN_TIMEOUT;
	  //okay, lets swap the new& old session.
	  newSession->callbacks=ThisSession->callbacks; //move the callbacks
	  ThisSession->callbacks=0; //but we don't want them zeroed on free.
	  l_item->data=newSession; //set the current list item to the new struct
	  FreeSession(ThisSession); //free the old session
	  ThisSession=newSession; // and set it to the newsession.	  
	  newSession=0; //and avoid confusion.
	  arr_obj = MakeSessionArray(interp, ThisSession);

	  if (arr_obj) {
	    //Do 'after 1 callback_cmd changesList'
	    cmd_objv[0]=Tcl_NewStringObj("after", strlen("after"));
	    cmd_objv[1]=Tcl_NewIntObj(1);
	    cmd_objv[3]=arr_obj;
	    //loop over all callbacks:
	    for (cb_iter=0; cb_iter < g_list_length(ThisSession->callbacks); cb_iter++) {
	      cb_item = g_list_nth(ThisSession->callbacks, cb_iter);
	      if (cb_item->data) {	
		cmd_objv[2]=Tcl_NewStringObj(cb_item->data, strlen(cb_item->data));
		Tcl_EvalObjv(interp, 4, cmd_objv, TCL_EVAL_GLOBAL);
	      }	      
	    }
	  } else {
	    Tcl_BackgroundError(interp);
	  }

	} else {
	  //no changes, discard the new session data.
	  FreeSession(newSession);
	  newSession=0; //no longer valid.
	}


      } // if new session

    } // have callbacks

  } // for each state.
  
  //And restart the timer
  if (ev->num_callbacks > 0 && !ev->have_timer) {
    StartTimer(interp, ev);
  }

  return TCL_OK;
  
}

