/*  XmmsTcl - an xmms remote control library for tcl.
 *
 *  Copyright (C) 2002  Pascal Scheffers <pascal@scheffers.net>
 *
 *    $Id: xmmstcl.h,v 1.2 2002/03/16 23:27:52 pascal Exp $
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * External Interfaces
 */

#ifndef XMMSTCL_H
#define XMMSTCL_H

int XmmsCmd(ClientData clientData, 
	    Tcl_Interp *interp,
	    int objc, Tcl_Obj *CONST objv[]);

void XmmsCmdDelete(ClientData clientData);

/* 
 * the functions below this line are called from the XmmsCmd procedure above
 * and contain the implementations of the xmms "command"s
 *
 * These are all stateless at the moment, but there might come a day when 
 * state is carried from one call to the next, especially for queueing and 
 * event handler functings (do not exist yet). 
 */

int Xmms_IsRunning(Tcl_Interp *interp, int session);
int Xmms_Run(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
int Xmms_GetVersion(Tcl_Interp *interp, int session);
int Xmms_Play(Tcl_Interp *interp, int session);
int Xmms_Stop(Tcl_Interp *interp, int session);
int Xmms_Pause(Tcl_Interp *interp, int session);
int Xmms_Next(Tcl_Interp *interp, int session);
int Xmms_Eject(Tcl_Interp *interp, int session);
int Xmms_Quit(Tcl_Interp *interp, int session);
int Xmms_Previous(Tcl_Interp *interp, int session);
int Xmms_IsPlaying(Tcl_Interp *interp, int session);
int Xmms_IsPaused(Tcl_Interp *interp, int session);
int Xmms_Playlist(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session, int enqueue);
int Xmms_PlaylistPos(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session);
int Xmms_PlaylistDelete(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session);
int Xmms_Length(Tcl_Interp *interp, int session);
int Xmms_Clear(Tcl_Interp *interp, int session);
int Xmms_Time(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_Duration(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_Volume(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_MainVolume(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_Balance(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_Info(Tcl_Interp *interp, int session);
int Xmms_Repeat(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_Shuffle(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_File(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_Title(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);

int Xmms_MainWindow(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_PLWindow(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_EQWindow(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_ShowPreferences(Tcl_Interp *interp, int session);

int Xmms_AlwaysOnTop(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_AddUrl(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], int session);
int Xmms_EQ(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session);
int Xmms_EQPreamp(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session);
int Xmms_EQBand(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session);
int Xmms_Skin(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[],int session);

#endif /* XMMSTCL_H */
